/*
 * This Software (C) Copyright David Goodwin, 2008, 2009.
 * 
 *   This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package iasel.gui;

import iasel.*;
import iasel.interfaces.StatusMonitor;
import java.awt.geom.*;

import javax.swing.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

import java.io.File;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;
import javax.imageio.ImageIO;

/**
 *
 * @author  David
 */
public class ImageCalcPanel extends JPanel {
    private BufferedImage bImage = null;  // Background image
    private BufferedImage cImage = null;  // Drawn image
    private BufferedImage lImage = null;  // other image

    private int mode;
    
    protected double translation_x = 0;
    protected double translation_y = 0;
    protected double scale = 1.0;
    protected double scale_max = 3.0;
    protected double scale_min = 0.2;    
    protected boolean translating = false;
    //protected BitSet[] mask = null;
    
    // Provides feedback to some monitoring class
    protected StatusMonitor mon = null;
    
    // Calculation calibration stuff
    protected int topLeftX = -1;
    protected int topLeftY = -1;
    protected int bottomRightX = -1;
    protected int bottomRightY = -1;
    
    // Modes:
    public final static int MODE_NOOP = 0;
    public final static int MODE_DRAW = 1;
    public final static int MODE_FILL = 2;
    public final static int MODE_CALIBRATE = 3;
    
    public boolean calibrateDone = false;
    
    private String mTitle;
    private String mMessage;
    private boolean mShow;
    
    private MouseHandler mouse;
    
    public void reset() {
        bImage = null;
        cImage = null;
        lImage = null;
        mode = MODE_NOOP;
        translation_x = 0;
        translation_y = 0;
        scale = 1.0;
        scale_max = 3.0;
        scale_min = 0.2;
        translating = false;
        //mask = null;
        topLeftX = -1;
        topLeftY = -1;
        bottomRightX = -1;
        bottomRightY = -1;
        mTitle = "";
        mMessage = "";
        mShow = false;
        calibrateDone = false;
                
        mouse.xold = 0;
        mouse.yold = 0;
        mouse.last_x = -1;
        mouse.last_y = -1;
        mouse.firstx = -1;
        mouse.firsty = -1;
        mouse.drawing = false;
    }    
    
    /** Class constructor. Sets up everything ready for use.
     */
    public ImageCalcPanel() {
        initComponents();
        
        mouse = new MouseHandler();

        addMouseListener (mouse);
        addMouseMotionListener (mouse);
        addMouseWheelListener (mouse);
    }
    
    public void setMonitor(StatusMonitor m) {
        mon = m;
    }
    
    /** Paints the component including the background image, selections, etc.
     * 
     * @param g The graphics object to draw on.
     */
    @Override
    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        
        Graphics2D g2d = (Graphics2D)g;
        
        AffineTransform old_transform = g2d.getTransform();

        // scale and translate the image
        g2d.translate (translation_x, translation_y);
        g2d.scale (scale, scale);

        // First try to draw the image
        if (bImage != null) {
            // Draw the background image
            g2d.drawImage(bImage,0,0,this);

            // Draw the filled-selection image
            g2d.drawImage(cImage,0,0,this);
        }
        
        // Draw the selection-only image (if it has been generated)
        if (lImage != null) {
            g2d.drawImage(lImage,0,0,this);
        }
        
        // Draw the frame box
        g2d.setColor(Color.BLUE);
        g2d.drawRect(topLeftX, topLeftY, bottomRightX - topLeftX, bottomRightY - topLeftY);
 
        // Set the image transform
        g2d.setTransform (old_transform);        
        
        // Show a message if there is one
        if (mShow)
            drawMessage(mTitle, mMessage, g);
        
        // Now draw any mode-specific stuff.
        if (mode == MODE_CALIBRATE) {
            // Display calibration instructions.
            
            if (topLeftX == -1 && topLeftY == -1)
                drawMessage("Calibration: Item Boundary Selection", 
                        "Please click just inside the TOP " +
                        "LEFT corner of the item.", g);
            else
                drawMessage("Calibration: Item Boundary Selection", 
                        "Please click just inside the " +
                        "BOTTOM RIGHT corner of the item.", g);
        }
    }
    
    public void showMessage(String title, String message) {
        mShow = true;
        mTitle = title;
        mMessage = message;
    }
    
    public void hideMessage() {
        mShow = false;
    }
    
    private void drawMessage(String title, String message, Graphics g) {
        int hPos = this.getWidth() / 2;
        int vPos = this.getHeight() / 2;
        vPos -= 62.5;
        hPos -= 200;
        g.setColor(Color.WHITE);
        g.fillRect(hPos, vPos, 400, 125);
        g.setColor(Color.BLACK);
        g.drawRect(hPos, vPos, 400, 125);

        // Now draw the title bar:
        g.setColor(Color.BLUE);
        g.fillRect(hPos, vPos, 400, 19);
        g.setColor(Color.BLACK);
        g.drawRect(hPos, vPos, 400, 19);

        // Now draw the title text:
        Font theFont = g.getFont();
        g.setColor(Color.WHITE);
        g.setFont(theFont.deriveFont(Font.BOLD));
        g.drawString(title, hPos + 5, vPos + 15);
        g.setFont(theFont);

        // And now for the message text. This is a bit more complicated as
        // we have to wrap the text.
        g.setColor(Color.BLACK);
        if (message.length() <= 69)
            g.drawString(message ,hPos+15, vPos+35);
        else {
            // We have to split it every 45 characters unless we can find
            // a space within 10 characters in which case we will split there.
            String[] lines = wrapString(message, 65);
            hPos += 15;
            vPos += 35;
            for(String s:lines) {
                if (s != null) {
                    g.drawString(s, hPos, vPos);
                    vPos += 15;
                }
            }
        }
    }
   
    /** A function used by the code for displaying a message in the rendering widget.
     * Attempts to wrap a string to a specified column count. It will try to insert
     * line breaks between words.
     * @param text the string to wrap
     * @param column The number of columns to try and wrap to
     * @return The wrapped string, one line per array item.
     */
    protected String[] wrapString(String text, int column) {
        ArrayList<String> lines = new ArrayList<String>();
        
        String[] words = text.split(" ");
        int currentLineLength = 0;
        StringBuilder sb = new StringBuilder();
        for (String s:words) {
            if ((currentLineLength + s.length() + 1) <= column){
                sb.append(s + " ");
                currentLineLength += s.length() + 1;
            }
            else if (currentLineLength > 0) {
                lines.add(sb.toString());
                sb = new StringBuilder();
                currentLineLength = 0;
                sb.append(s + " ");
                currentLineLength += s.length() + 1;
            }
            else {
                sb.append(s + " ");
                currentLineLength += s.length() + 1;
            }
        }
        if (!sb.toString().equals("")) {
            lines.add(sb.toString());
        }
               
        return lines.toArray(new String[1]);
    }    
    
    /** Runs an image process. This involves opening the image, displaying
     * it, getting the user to select the dimensions, getting the user to draw
     * the selection and finally calculating the selection
     * 
     * @param imageFile the image file to display
     */    
    public void startProcess(File imageFile) {
        try {
            if (mon != null) mon.busy(true);
           
            // Load the image.
            if (mon != null) mon.statusMessage("Loading " + imageFile.getAbsolutePath() + "...");
           
            bImage = ImageIO.read(imageFile);
            System.out.println("Loaded " + imageFile.getAbsolutePath());

            // Clear the selection (in this case initialise the structures)
            if (mon != null) mon.statusMessage("Clearing Selections...");
            clearSelection();            
            
            // repaint.
            repaint();
            if (mon != null) mon.statusMessage("");
            if (mon != null) mon.busy(false);
        } catch (Exception e) {
            //@TODO: There was an error. Tell the user.
        }
    }
    
    /** Clears the selection by recreating all the data structures, then
     * repaints the component.
     */
    public void clearSelection() {
        // If there is no image loaded, there cant be any selection to clear.
        if (bImage == null) return;
        
        // create a new cImage.
        //cImage = new BufferedImage(bImage.getWidth(), bImage.getHeight(), IndexColorModel.BITMASK);
        cImage = null;
        cImage = new BufferedImage(bImage.getWidth(), bImage.getHeight(), BufferedImage.TYPE_INT_ARGB);
        Graphics g = cImage.getGraphics();
        g.setColor(new Color(0,0,0,0));
        g.drawRect(0,0,bImage.getWidth(), bImage.getHeight());

        // Initialise the bit mask thingey
        /*mask = new BitSet[bImage.getHeight()];
        for (int i = 0; i < bImage.getHeight(); i++) {
            mask[i] = new BitSet(bImage.getWidth());
        } */       
        lImage = null;
        repaint();
    }

    /** Sets the mode. Use mode constants such as MODE_NOOP, MODE_DRAW, 
     * MODE_FILL, etc.
     * 
     * @param mode The mode to set it to.
     */
    public void setMode(int mode) {
        
        if (bImage == null && mode != MODE_NOOP) {
            // We cant change modes when there is no image loaded. Set the mode
            // to MODE_NOOP and notify the parent of the change.
            this.mode = MODE_NOOP;
            mon.ModeChanging(MODE_NOOP);
            return;
        }
        
        // Store the mode
        this.mode = mode;
        
        if (this.mode == MODE_CALIBRATE) {
            topLeftX = -1;
            topLeftY = -1;
            bottomRightX = -1;
            bottomRightY = -1;
        }
            
        
        // Then repaint incase this mode needs to display instructions.
        repaint();
    }
    
    /** This notifies the monitor that the mode is changing and then changes it.
     * 
     * @param mode The mode to switch to.
     */
    private void switchMode(int mode) {
        mon.ModeChanging(mode);
        setMode(mode);        
    }
    
    /** Gets the selection percentage.
     * 
     * @return
     */
    public double getPercentage() {
        double framePixels = countFramePixels();
        double selectionPixels = countSelectionPixels();
        return (selectionPixels/framePixels) * 100;
    }
    
    /** Counts the number of pixels in the frame
     * @return Number of pixels in the frame
     */
    public int countFramePixels() {
        int rows = topLeftY - bottomRightY;
        int columns = topLeftX - bottomRightX;
        return rows * columns;
    }
    
    /** Counts the number of selected (red) pixels.
     * 
     * @return Number of selected pixels
     */
    public int countSelectionPixels() {
        int count = 0;
        for (int x = 0; x < bImage.getWidth(); x++) {
            for (int y = 0; y < bImage.getHeight(); y++) {
                if (getPoint(x, y)) count++;
            }
        }
        
        return count;
    }
    
    /** Maps an X coordinate such as those given my the mouse pointer to the
     * translated and scaled coordinates used in the drawing area. 
     * 
     * @param x The X Coordinate to scale and translate
     * @return Translated and scaled coordinate
     */
    int mapx (int x){
        return (int)((x - translation_x) / scale);
    }

    /** Maps an Y coordinate such as those given my the mouse pointer to the
     * translated and scaled coordinates used in the drawing area. 
     * 
     * @param y The Y Coordinate to scale and translate
     * @return Translated and scaled coordinate
     */    
    int mapy (int y){
        return (int)((y - translation_y) / scale);
    }
 
    /** Sets a selected point in all appropriate data structures (cImage for
     * drawing to the user, the mask for calculating).
     * 
     * @param x X position
     * @param y Y position
     */
    void setPoint(int x, int y) {
        //mask[y].set(x);
        
        cImage.setRGB(x,y, Color.RED.getRGB());
        
                
        //Graphics2D g2d = (Graphics2D)(cImage.getGraphics());
        
        //g2d.setColor(Color.RED);
        //g2d.drawRect(x, y, 1, 1);

    }
    
    /** Gets a point from the selection bitmap
     * 
     * @param x X coordinate
     * @param y Y coordinate
     * @return The value (set or unset, true or false)
     */
    boolean getPoint(int x, int y) {
        //return mask[y].get(x);
        int i = cImage.getRGB(x, y);
        //System.out.println(Integer.toString(i));
        if (i == Color.RED.getRGB()) return true;
        return false;
    }
    
    /** Generates sImage, the selection image (in blue), from the mask data
     * structure. This is mostly just used for testing/debugging.
     */
    void makeSelectionImage() {
        lImage = new BufferedImage(bImage.getWidth(), bImage.getHeight(), IndexColorModel.BITMASK);
        Graphics2D g2d = (Graphics2D)(lImage.getGraphics());
        g2d.setColor(Color.BLUE);
        
       /* for (int y = 0; y < mask.length; y++) {
            for (int x = 0; x < bImage.getWidth(); x++) {
                if (getPoint(x,y))
                    g2d.drawRect(x,y,1,1);
            }
        }*/
    }
    
    private class coordinate {
        /** Constructs the class
         * 
         * @param x_value The initial X value
         * @param y_value The initial Y value
         */
        public coordinate(int x_value, int y_value) {
            x = x_value; y = y_value;
        }
        
        public coordinate(coordinate c) {
            x = c.x;
            y = c.y;
        }
        
        /** Moves this coordinate one point to the left.
         */
        public coordinate getWest() {
            return new coordinate(x-1,y);
        }
        
        /** Moves this coordinate one point to the right.
         */
        public coordinate getEast() {
            return new coordinate(x+1,y);
        }
        
        public coordinate getNorth() {
            return new coordinate(x,y+1);
        }
        
        public coordinate getSouth() {
            return new coordinate(x,y-1);
        }
        
        public boolean getValue() {
            return getPoint(x,y);
        }
        
        public void setRed() {
            setPoint(x,y);
        }
        
        @Override
        public String toString() {
            return x + ", " + y;
        }
        
        @Override
        public int hashCode() {
            return x | y << 16;
        }
        
        @Override
        public boolean equals (Object o) {
            if (o == null) return false;
            if (! (o instanceof coordinate)) return false;

            coordinate c = (coordinate)o;
            return c.x == x && c.y == y;
        }
        
        int x;
        int y;
    }
    
    /** Performs a flood fill inside a selected area. The target colour is 0
     * (unselected) and the replacmenet colour is 1 (selected). It should not
     * be used outside a selected area.
     * 
     * @param x The X coordinate to start filling at.
     * @param y The Y coordinate to start filling at.
     */
    void flood(int x, int y) {
        flood1(x,y);
    }
    
    void flood2(int x, int y) {
        // DEBUG
        System.out.println("Staring flood fill at x: " + x + " Y: " + y);
        
        // Create a new empty queue. This is what we will be working from.
        LinkedList<coordinate> queue = new LinkedList<coordinate>();
        Set<coordinate> setThing = new HashSet<coordinate>();
        
        // If the specified point is RED then we dont have anything more to do.
        // We might as well just return rather than waste time doing nothing.
        coordinate start = new coordinate(x,y);
        
        if (start.getValue()) return;
        
        // Now we add the specified coordinates to the queue
        queue.add(start);
        
        // Loop for as long as there are things in the queue.
        while (!queue.isEmpty()) {
            // grab the node from the start of the queue
            coordinate c = queue.removeFirst();
            
            // Find all the successor nodes and add them to the queue. This will
            // also colour all pixels on the current row.
            ArrayList aa = successorRun(c,setThing);
            queue.addAll(aa);
            setThing.addAll(aa);
            //System.out.println(queue.size());
        }
        

        // and then we repaint
        repaint();
    }
    
    ArrayList<coordinate> successorStep(coordinate c, Set<coordinate> s) {
        ArrayList<coordinate> a = new ArrayList<coordinate>();
        
        // Get all the nodes around the current one
        coordinate north = c.getNorth();
        coordinate south = c.getSouth();
        coordinate east = c.getEast();
        coordinate west = c.getWest();
        
        // For each surrounding node, add it to the queue if it is not red.
        if (!north.getValue() && !s.contains(north)) a.add(north);
        if (!south.getValue() && !s.contains(south)) a.add(south);
        if (!east.getValue() && !s.contains(east)) a.add(east);
        if (!west.getValue() && !s.contains(west)) a.add(west);
        
        // Set the current nodes value to be red
        c.setRed();
        
        // Then return the list of nodes that still need to be visited.
        return a;
    }
    
    ArrayList<coordinate> successorRun(coordinate c, Set<coordinate> s) {
        ArrayList<coordinate> a = new ArrayList<coordinate>();
        
        coordinate west = new coordinate(c);
        coordinate east = new coordinate(c);
        coordinate north = null;
        coordinate south = null;
        
        // First we need to walk west as far as we can.
        while (!west.getValue()) {
            // Grab the nodes to the north and south of this one.
            north = west.getNorth();
            south = west.getSouth();
            
            // If the one to the north isnt red, add it to the REDification queue
            if (!north.getValue()) a.add(north);
            
            // Do the same for the one to the south
            if (!south.getValue()) a.add(south);
            
            // Colour the current node.
            west.setRed();
            
            // And then do one step to the west.
            west = west.getWest();
        }
        
        // Now we do the same for the east
        while (!east.getValue()) {
            // Grab the nodes to the north and south of this one.
            north = east.getNorth();
            south = east.getSouth();
            
            // If the one to the north isnt red, add it to the REDification queue
            if (!north.getValue()) a.add(north);
            
            // Do the same for the one to the south
            if (!south.getValue()) a.add(south);
            
            // Colour the current node.
            east.setRed();            
            
            // And then do one step to the east.
            east = east.getEast();
        }        
        
        // And now we just have to draw a line between west and east to fill
        // all the nodes we have walked along today.
        //drawLine(west.x,west.y,east.x, east.y);
        //System.out.println("East: " + east.toString() + " West: " + west.toString());
        
        return a;
    }
    
    void flood1(int x, int y) {
        final boolean target = false;
        
        if (mon != null) mon.statusMessage("Performing Flood Fill");
        if (mon != null) mon.busy(true);


        // Create a new queue of vertical nodes to deal with
        LinkedList<Integer> xQueue = new LinkedList<Integer>();
        LinkedList<Integer> yQueue = new LinkedList<Integer>();
        
        // If the selected node isnt a target then just return.
        if (getPoint(x,y) != target) return;

        // Add the current node to the queue.
        System.out.println("  Adding intitial node to queue");
        xQueue.add(x);
        yQueue.add(y);

        // Loop for as long as there is something in the queue
        while (!xQueue.isEmpty()) {
            // Grab the current item from the queue
            int xx = xQueue.removeFirst();
            int yy = yQueue.removeFirst();       

            // If the current item is a target, it is time to do some filling
            if (getPoint(xx,yy) == target) {

                // We will be moving east and west in these variables.
                int xwest = xx;
                int xeast = xx;
                
                // Keep moving west until we hit a pixel that isnt a target
                while ((getPoint(xwest,yy) == target) && (xwest >= 0)) {

                    // If the pixel to the north is a target, stick it in the
                    // queue
                    if (yy+1 < bImage.getHeight())//mask.length)
                        if (getPoint(xwest,yy+1) == target) {
                            xQueue.add(xwest);
                            yQueue.add(yy+1);
                        }
                    // If the pixel to the south is a target, stick it in the
                    // queue.
                    if (yy-1 >= 0)
                        if (getPoint(xwest,yy-1) == target) {
                            xQueue.add(xwest);
                            yQueue.add(yy-1);
                        }                 
                    
                    setPoint(xwest,yy);
                    
                    // Continue west
                    xwest--;
                    if (xwest < 0) break;
                }
 
                xeast++;
                // Keep moving east until we hit a pixel that isnt a target               
                
                // Because the while loops condition contains a call to get the
                // value at a point, we must make sure that the initial values
                // are valid. If they arent, as there is nothing beyond this
                // loop we can just continue.
                //
                //if (xeast >= bImage.getWidth()) continue;
                //if (yy >= bImage.getHeight()) continue;
                
                while ((getPoint(xeast,yy) == target) && (xeast < bImage.getWidth())) {

                    // If the pixel to the north is a target, stick it in the
                    // queue.
                    if (yy+1 < bImage.getHeight())//mask.length)
                        if (getPoint(xeast,yy+1) == target) {
                            xQueue.add(xeast);
                            yQueue.add(yy+1);
                        }
                    // If the pixel to the south is a target, stick it in the
                    // queue.
                    if (yy-1 >= 0)
                        if (getPoint(xeast,yy-1) == target) {
                            xQueue.add(xeast);
                            yQueue.add(yy-1);
                        }   
                    
                    setPoint(xeast,yy);
                    
                    // And continue east.
                    xeast++;
                    
                    // Check the coordinates:
                    if (xeast >= bImage.getWidth()) break;
                    if (yy >= bImage.getHeight()) break;
                }
            }
            // Then continue looping: grab the next item from the queue
            // and do it all again.
        }
        
        // Repaint the display then return
        repaint();
        if (mon != null) mon.busy(false);
        if (mon != null) mon.statusMessage("");
    }
    
    /** Draws a nice line with no nasty gaps into the selection data structures.
     * 
     * @param x0 Start coordinate X value.
     * @param y0 Start coordinate Y value.
     * @param x1 End coordinate X value.
     * @param y1 End coordinate Y value.
     */
    private void drawLine(int x0, int y0, int x1, int y1) {
        int dx = x1 - x0;
        int dy = y1 - y0;

        setPoint(x0,y0);
        if (Math.abs(dx) > Math.abs(dy)) {          // slope < 1
            float m = (float) dy / (float) dx;      // compute slope
            float b = y0 - m*x0;
            dx = (dx < 0) ? -1 : 1;
            while (x0 != x1) {
                x0 += dx;
                setPoint(x0,Math.round(m*x0 + b));
            }
        } else
        if (dy != 0) {                              // slope >= 1
            float m = (float) dx / (float) dy;      // compute slope
            float b = x0 - m*y0;
            dy = (dy < 0) ? -1 : 1;
            while (y0 != y1) {
                y0 += dy;
                setPoint(Math.round(m*y0 + b),y0);
            }
        }
    }    
    
    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        org.jdesktop.application.ResourceMap resourceMap = org.jdesktop.application.Application.getInstance(iasel.Iasel.class).getContext().getResourceMap(ImageCalcPanel.class);
        setBackground(resourceMap.getColor("Form.background")); // NOI18N
        setName("Form"); // NOI18N

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 400, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 300, Short.MAX_VALUE)
        );
    }// </editor-fold>//GEN-END:initComponents
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    // End of variables declaration//GEN-END:variables
 
    /** The Mouse Handler handles all mouse input. It is very important.
     */
    class MouseHandler extends MouseInputAdapter {

        public int xold = 0;
        public int yold = 0; 
        
        // Use these to join the dots
        public int last_x = -1, last_y = -1;
        public int firstx = -1, firsty = -1;
        
        public boolean drawing = false;
        
        /** Performs the appropriate action when the mouse is clicked. What the
         * appropriate action is depends on what mode the ImageCalcPanel is in.
         * 
         * @param e The MouseEvent to process.
         */
        @Override
        public void mousePressed (MouseEvent e){
            // Before we proceed, check that we actually have an image open.
            // bImage is the background image. If this is still null, nothing
            // has been loaded yet, so we might as well just do nothing and
            // return.
            if (bImage == null)
                return;
            
            
            // Set the old coordinates.
            xold = e.getX();
            yold = e.getY();    
            
            System.out.println("X: " + mapx(e.getX()) + " Y: " + mapy(e.getY()) + " V: " + getPoint(mapx(e.getX()),mapy(e.getY())));
            
            // We right-click drag to move the image around
            if (e.getButton() == e.BUTTON3){
                translating = true;
            } 
            // If we left click then what we do depends on the mode
            if (e.getButton() == e.BUTTON1) {
                // draw a selection
                if (mode == MODE_DRAW)
                    drawing = true;
                // Fill a selection
                else if (mode == MODE_FILL) {
                    flood(mapx(e.getX()),mapy(e.getY()));
                }
                else if (mode == MODE_CALIBRATE) {
                    if (topLeftX == -1 && topLeftY == -1) {
                        // We are setting the top left corner.
                        topLeftX = mapx(e.getX());
                        topLeftY = mapy(e.getY());
                    } else {
                        bottomRightX = mapx(e.getX());
                        bottomRightY = mapy(e.getY());
                        switchMode(MODE_DRAW);
                        calibrateDone = true;
                    }
                    repaint();
                }
            }
        }

        /** The user has finished dragging or clicking. Stop 
         * drawing/translating/whatever and repaint.
         * 
         * @param e The mouse event.
         */
        @Override
        public void mouseReleased (MouseEvent e){
          if (drawing && translating) {
              // Finish the line
              drawLine(firstx, firsty, last_x, last_y);
              repaint();
          }
          drawing = false;
          translating = false;
          
        last_x = -1;
        last_y = -1;          
        firstx = -1;
        firsty = -1;        
        }

        /** The user is dragging. Translte, draw, or whatever.
         * 
         * @param e The mouse event.
         */
        @Override
        public void mouseDragged (MouseEvent e){
          int xrel = e.getX() - xold;
          int yrel = e.getY() - yold;

          xold = e.getX();
          yold = e.getY();          
          boolean paint = false;

          if (translating && !drawing){
              translation_x += xrel;
              translation_y += yrel;
              paint = true;
          } 

          if (drawing && bImage != null) {
              int x,y;

              int mx = mapx(e.getX());
              int my = mapy(e.getY());

              if (my > bImage.getHeight()) y = bImage.getHeight() -1;
              else if (my < 0) y = 0;
              else y = my;

              if (mx > bImage.getWidth()) x = bImage.getWidth() -1;
              else if (mx < 0) x = 0;
              else x = mx;


              // set all the points between last_x,last_y and x,y

              //mask[y].set(x);
              if (last_x == -1 || last_y == -1) {
                  last_x = x;
                  last_y = y;
              }
              if (firstx == -1 || firsty == -1) {
                  firstx = x;
                  firsty = y;
              }
              drawLine(x,y,last_x,last_y);

              last_x = x;
              last_y = y;

              paint = true;
            }

          if (paint)
            repaint();
        }
      
      @Override
      public void mouseMoved (MouseEvent e){

      }

      @Override
      public void mouseWheelMoved (MouseWheelEvent e) {
            int notches = e.getWheelRotation();

            double cx1 = mapx (e.getX());
            double cy1 = mapy (e.getY());

            scale += 0.05 * notches;

            //Limit the scaling
            if (scale > scale_max)
            scale = scale_max;
            if (scale < scale_min)
            scale = scale_min;

            //Do some translation correction.
            double cx2 = mapx (e.getX());
            double cy2 = mapy (e.getY());
            translation_x += (cx2 - cx1) * scale;
            translation_y += (cy2 - cy1) * scale;

            repaint();
        }
    }    
}
